/**
 *   Copyright (C) 2009  Cedric Tabin
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef PRETTY_PRINTER_MAIN_H
#define PRETTY_PRINTER_MAIN_H

//========================================== INCLUDES =========================================================

#ifdef HAVE_LIBXML
#include <libxml/parser.h>
#include <libxml/tree.h>
#endif

#include "PrettyPrinter.h"

//========================================== PARAMETERS =======================================================

bool checkXML = TRUE; //if the input must be checked as XML
PrettyPrintingOptions* ppOptions = NULL; //options for the pretty-printing

//========================================== FUNCTIONS ========================================================

void showUsage();
int parseArgs(FILE* err, int argc, const char* argv[]);
int parseLongArg(FILE* err, int argc, const char* argv[], int* currentArg, const char* longValue, const char* shortValue);
char* readInput(FILE* in);
void writeOutput(FILE* out, const char* output);
void initializeOptions(PrettyPrintingOptions* ppOptions);
char convertEscapedChars(const char *a);

#ifdef HAVE_LIBXML
int checkXMLInput(const char* input);
#endif

//========================================== IMPLEMENTATION ===================================================

int main(int argc, const char* argv[])
{
    //initialization of constants
    FILE* in = stdin;
    FILE* out = stdout;
    FILE* err = stderr;
    
    ppOptions = createDefaultPrettyPrintingOptions();
    ppOptions->oneLineText = FALSE;
    ppOptions->inlineText = FALSE;
    ppOptions->oneLineComment = FALSE;
    ppOptions->inlineComment = FALSE;
    ppOptions->oneLineCdata = FALSE;
    ppOptions->inlineCdata = FALSE;
    ppOptions->emptyNodeStripping = FALSE;
    ppOptions->emptyNodeStrippingSpace = FALSE;
    ppOptions->forceEmptyNodeSplit = FALSE;
    ppOptions->trimLeadingWhites = FALSE;
    ppOptions->trimTrailingWhites = FALSE;
    ppOptions->alignComment = FALSE;
    ppOptions->alignText = FALSE;
    ppOptions->alignCdata = FALSE;
  
    //parse args
    int result = parseArgs(err, argc, argv);
    if (result!=0)
    {
        showUsage();
        return result;
    }
  
    //read and check if the input is xml-valid
    char* input = readInput(in);
    if(strlen(input) == 0) { fprintf(err, "Error: input is empty\n"); return 1; }
    #ifdef HAVE_LIBXML
    if(checkXML && checkXMLInput(input) != 0)
    {
        fprintf(err, "Error: input is not XML-valid\n");
        return 1;
    }
    #endif
  
    //pretty printing
    int length = strlen(input);
    int ppResult = processXMLPrettyPrinting(&input, &length, ppOptions);
    if (ppResult != 0)
    {
        fprintf(err, "Error: unable to process the xml pretty printing (return code is %d)\n", ppResult);
        return ppResult;
    }
  
    //write the output
    writeOutput(out, input);
    
    //free all what we used
    free(ppOptions);
    free(input);
    
    return 0;
}

void showUsage()
{
    fprintf(stdout, "Usage: xml-pretty-print [OPTIONS]\n\n");
    fprintf(stdout, "Options are:\n");
    fprintf(stdout, "  --help (-h)                   show this help\n");
    fprintf(stdout, "  --version (-v)                shows the version and exit\n");
#ifdef HAVE_LIBXML
    fprintf(stdout, "  --no-check                    does not check if the input is a valid XML (the pretty-printing may fail)\n");
#endif
    fprintf(stdout, "  --default                     uses the default configuration\n");
    fprintf(stdout, "  --compact                     uses a compact configuration\n");
    fprintf(stdout, "  --single-line                 put the xml on one line (no pretty printing)\n");
    fprintf(stdout, "  --one-line-text (-t)          put the text nodes on one line\n");
    fprintf(stdout, "  --inline-text (-T)            inline text nodes (if possible)\n");
    fprintf(stdout, "  --align-text (-g)             align the text nodes (only if -t is not set)\n");
    fprintf(stdout, "  --one-line-comment (-o)       put the comments on one line\n");
    fprintf(stdout, "  --inline-comment (-i)         inline comments (if possible)\n");
    fprintf(stdout, "  --align-comment (-a)          align the comments (only if -c is not set)\n");
    fprintf(stdout, "  --one-line-cdata (-O)         put the cdata nodes on one line\n");
    fprintf(stdout, "  --inline-cdata (-I)           inline cdata nodes (if possible)\n");
    fprintf(stdout, "  --align-cdata (-A)            align the cdata (only if -C is not set)\n");
    fprintf(stdout, "  --trim (-r)                   trim leading and trailing spaces in text nodes\n");
    fprintf(stdout, "  --empty-node-strip (-s)       concat <x></x> to <x/>\n");
    fprintf(stdout, "  --empty-node-space (-S)       expand <x/> to <x /> (add a space)\n");
    fprintf(stdout, "  --empty-node-expand (-x)      expand <x/> to <x></x> (only if -s is not set)\n");
    fprintf(stdout, "  --indent-length (-l) LENGTH   number of chars (-c) to set at each level of indentation\n");
    fprintf(stdout, "  --indent-char (-c) CHAR       char to use for the indentation (supports \\t, \\n, \\r, \\b, \\v and any other simple char)\n");
    fprintf(stdout, "  --line-break (-n) TYPE        type of line break. The allowed values are LINUX (\\n), MAC (\\r) or WINDOWS (\\r\\n)\n");
    
    fprintf(stdout, "\nCopyright (c) 2011\n");
    fprintf(stdout, "Cedric Tabin - http://www.astorm.ch\n");
}

int parseArgs(FILE* err, int argc, const char* argv[])
{
    int counter = 1;
    while(counter < argc)
    {
        const char* arg = argv[counter];
        int arglen = strlen(arg);
        if(arglen < 2 || arg[0] != '-') { fprintf(err, "Invalid option: '%s'", arg); return 1; }
        
        bool isSingleArg = (arg[1] == '-'); //argument begins with '--'
        
        if (isSingleArg)
        {
            int result = parseLongArg(err, argc, argv, &counter, arg, NULL);
            if (result != 0) { return result; } //exit because error
        }
        else
        {
            int localCounter = 1;
            while(localCounter < arglen)
            {
                char c = arg[localCounter];
                int r = 0;
                if (c == 'h') { r = parseLongArg(err, argc, argv, &counter, "--help", "-h"); }

                else if (c == 't') { r = parseLongArg(err, argc, argv, &counter, "--one-line-text", "-t"); }
                else if (c == 'T') { r = parseLongArg(err, argc, argv, &counter, "--inline-text", "-T"); }
                else if (c == 'g') { r = parseLongArg(err, argc, argv, &counter, "--align-text", "-g"); }
                else if (c == 'o') { r = parseLongArg(err, argc, argv, &counter, "--one-line-comment", "-o"); }
                else if (c == 'i') { r = parseLongArg(err, argc, argv, &counter, "--inline-comment", "-i"); }
                else if (c == 'a') { r = parseLongArg(err, argc, argv, &counter, "--align-comment", "-a"); }
                else if (c == 'O') { r = parseLongArg(err, argc, argv, &counter, "--one-line-cdata", "-O"); }
                else if (c == 'I') { r = parseLongArg(err, argc, argv, &counter, "--inline-cdata", "-I"); }
                else if (c == 'A') { r = parseLongArg(err, argc, argv, &counter, "--align-cdata", "-A"); }
                else if (c == 'r') { r = parseLongArg(err, argc, argv, &counter, "--trim", "-r"); }
                else if (c == 's') { r = parseLongArg(err, argc, argv, &counter, "--empty-node-strip", "-s"); }
                else if (c == 'S') { r = parseLongArg(err, argc, argv, &counter, "--empty-node-space", "-S"); }
                else if (c == 'x') { r = parseLongArg(err, argc, argv, &counter, "--empty-node-expand", "-x"); }
                else if (c == 'l') { r = parseLongArg(err, argc, argv, &counter, "--indent-length", "-l"); }
                else if (c == 'c') { r = parseLongArg(err, argc, argv, &counter, "--indent-char", "-c"); }
                else if (c == 'n') { r = parseLongArg(err, argc, argv, &counter, "--line-break", "-n"); }
                else if (c == 'v') { r = parseLongArg(err, argc, argv, &counter, "--version", "-v"); }

                else { fprintf(err, "Single option not recognized: -%c\n", c); return 1; }

                if (r != 0) { return r; } //exit because error
                ++localCounter;
            }
        }
        
        ++counter;
    }
    
    return 0;
}

int parseLongArg(FILE* err, int argc, const char* argv[], int* currentArg, const char* arg, const char* opt)
{
    if (strncmp(arg, "--help", 6) == 0) { return 1; }
#ifdef HAVE_LIBXML
    else if(strncmp(arg, "--version", 9) == 0) { fprintf(stdout, "Full v%s\n", PRETTY_PRINTER_VERSION); exit(0); }
#else
    else if(strncmp(arg, "--version", 9) == 0) { fprintf(stdout, "Light v%s\n", PRETTY_PRINTER_VERSION); exit(0); }
#endif
    else if(strncmp(arg, "--no-check", 10) == 0) { checkXML = FALSE; }
    else if (strncmp(arg, "--single-line", 13) == 0)
    {
        ppOptions->newLineChars = "";
        ppOptions->indentChar = ' ';
        ppOptions->indentLength = 0;
        ppOptions->alignComment = FALSE;
        ppOptions->alignText = FALSE;
        ppOptions->alignCdata = FALSE;
        ppOptions->oneLineCdata = TRUE;
        ppOptions->oneLineComment = TRUE;
        ppOptions->oneLineText = TRUE;
    }
    else if (strncmp(arg, "--compact", 13) == 0)
    {
        ppOptions->alignComment = TRUE;
        ppOptions->alignText = TRUE;
        ppOptions->alignCdata = TRUE;
        ppOptions->oneLineCdata = TRUE;
        ppOptions->oneLineComment = TRUE;
        ppOptions->oneLineText = TRUE;
        ppOptions->inlineCdata = TRUE;
        ppOptions->inlineComment = TRUE;
        ppOptions->inlineText = TRUE;
        ppOptions->trimLeadingWhites = TRUE;
        ppOptions->trimTrailingWhites = TRUE;
        ppOptions->emptyNodeStripping = TRUE;
        ppOptions->emptyNodeStrippingSpace = FALSE;
    }
    else if (strncmp(arg, "--default", 9) == 0)
    {
        free(ppOptions);
        ppOptions = createDefaultPrettyPrintingOptions();
    }
    else if (strncmp(arg, "--one-line-text", 15) == 0) { ppOptions->oneLineText = TRUE; }
    else if (strncmp(arg, "--inline-text", 13) == 0) { ppOptions->inlineText = TRUE; }
    else if (strncmp(arg, "--one-line-comment", 18) == 0) { ppOptions->oneLineComment = TRUE; }
    else if (strncmp(arg, "--inline-comment", 16) == 0) { ppOptions->inlineComment = TRUE; }
    else if (strncmp(arg, "--one-line-cdata", 16) == 0) { ppOptions->oneLineCdata = TRUE; }
    else if (strncmp(arg, "--inline-cdata", 14) == 0) { ppOptions->inlineCdata = TRUE; }
    else if (strncmp(arg, "--align-cdata", 13) == 0) { ppOptions->alignCdata = TRUE; }
    else if (strncmp(arg, "--align-comment", 15) == 0) { ppOptions->alignComment = TRUE; }
    else if (strncmp(arg, "--align-text", 12) == 0) { ppOptions->alignText = TRUE; }
    else if (strncmp(arg, "--trim", 6) == 0)
    {
        ppOptions->trimLeadingWhites = TRUE;
        ppOptions->trimTrailingWhites = TRUE;
    }
    else if (strncmp(arg, "--empty-node-strip", 18) == 0) { ppOptions->emptyNodeStripping = TRUE; }
    else if (strncmp(arg, "--empty-node-space", 18) == 0) { ppOptions->emptyNodeStrippingSpace = TRUE; }
    else if (strncmp(arg, "--empty-node-expand", 19) == 0) { ppOptions->forceEmptyNodeSplit = TRUE; }
    else if (strncmp(arg, "--indent-length", 15) == 0) 
    { 
        if (*currentArg >= argc-1) { fprintf(err, "Integer argument needed for %s (%s)\n", arg, opt); return 1; }
        
        (*currentArg)++;
        const char* nbValue = argv[*currentArg];
        if (strlen(nbValue) == 0) { fprintf(err, "Invalid value for %s (%s)\n", arg, opt); return 1; }
        
        ppOptions->indentLength = atoi(nbValue);;
    }
    else if (strncmp(arg, "--indent-char", 13) == 0) 
    { 
        if (*currentArg >= argc-1) { fprintf(err, "Argument needed for %s (%s)\n", arg, opt); return 1; }
        
        (*currentArg)++;
        const char* charValue = argv[*currentArg];
        if (strlen(charValue) == 0 ||
            strlen(charValue) > 2) { fprintf(err, "Invalid value for %s (%s)\n", arg, opt); return 1; }
        
        char parsed = convertEscapedChars(charValue);
        ppOptions->indentChar = parsed;
    }
    else if (strncmp(arg, "--line-break", 12) == 0) 
    { 
        if (*currentArg >= argc-1) { fprintf(err, "Argument needed for %s (%s)\n", arg, opt); return 1; }
        
        (*currentArg)++;
        const char* type = argv[*currentArg];
        if (strncmp(type, "LINUX", 5) == 0) { ppOptions->newLineChars = "\n"; }
        else if (strncmp(type, "MAC", 3) == 0) { ppOptions->newLineChars = "\r"; }
        else if (strncmp(type, "WINDOWS", 7) == 0) { ppOptions->newLineChars = "\r\n"; }
        else { fprintf(err, "Invalid type '%s' for argument %s\n", type, arg); return 1; }
    }
    
    else { fprintf(err, "Option not recognized: %s\n", arg); return 1;  }
    
    return 0;
}

char* readInput(FILE* in)
{
    int capacity = 1000; //1kb - capacity will be doubled (1kb, 2kb, 4kb, ...) each time more space is needed
    char* input = malloc(sizeof(char)*capacity);
    
    int counter = 0;
    while(feof(in) == 0)
    {
        //need to readjust the size
        if (counter == capacity)
        {
            capacity *= 2;
            input = realloc(input, sizeof(char)*capacity);
        }
      
        //read the next char and put it in the input buffer
        char read;
        int nb = fread(&read, sizeof(char), 1, in);
        if (nb == 1)
        {
            input[counter] = read;
            ++counter;
        }
    }
    
    //adjust the final size of the input data and set the final char (guard)
    input = realloc(input, sizeof(char)*counter+1);
    input[counter] = '\0';
    
    return input;
}

void writeOutput(FILE* out, const char* output)
{
    fprintf(out, "%s", output);
}

#ifdef HAVE_LIBXML
int checkXMLInput(const char* input)
{
    //checks if the data is an XML format
    xmlDoc* parsedDocument = xmlParseDoc((const unsigned char*)input);

    //this is not a valid xml
    if(parsedDocument == NULL) { return 1; }
    
    xmlFreeDoc(parsedDocument);
    
    return 0;
}
#endif

/*
 * TODO: find a better solution (ie using iconv). However, the set of chars below
 * will answer most of queries.
 */
char convertEscapedChars(const char *a) 
{
    if (strlen(a) == 1) { return a[0]; }
    if (strncmp(a, "\\t", 2) == 0) { return '\t'; }
    if (strncmp(a, "\\n", 2) == 0) { return '\n'; }
    if (strncmp(a, "\\r", 2) == 0) { return '\r'; }
    if (strncmp(a, "\\b", 2) == 0) { return '\b'; }
    if (strncmp(a, "\\v", 2) == 0) { return '\v'; }
    
    return ' ';
}

#endif
